# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import time
import itertools
import unittest
import threading
import operator
import copy
import tempfile
import logging
import queue
import multiprocessing
from concurrent import futures
from functools import partial

import dimod
from parameterized.parameterized import parameterized

from hybrid.flow import (
    Branch, Branches, RacingBranches, ParallelBranches, Parallel,
    ArgMin, Map, Reduce, Lambda, Unwind, TrackMin,
    LoopUntilNoImprovement, LoopWhileNoImprovement, SimpleIterator, Loop,
    Identity, BlockingIdentity, Dup, Const, Wait, Log
)
from hybrid.core import State, States, Runnable, Present
from hybrid.utils import min_sample, max_sample
from hybrid.profiling import tictoc
from hybrid.exceptions import EndOfStream, StateDimensionalityError
from hybrid.testing import mock
from hybrid import traits


class TestBranch(unittest.TestCase):

    def test_empty(self):
        with self.assertRaises(ValueError):
            Branch()

    def test_composition(self):
        class A(Runnable):
            def next(self, state):
                return state.updated(x=state.x + 1)
        class B(Runnable):
            def next(self, state):
                return state.updated(x=state.x * 7)

        a, b = A(), B()
        s = State(x=1)

        b1 = Branch(components=(a, b))
        self.assertEqual(b1.components, (a, b))
        self.assertEqual(b1.run(s).result().x, (s.x + 1) * 7)

        b2 = b1 | b | a
        self.assertEqual(b2.components, (a, b, b, a))
        self.assertEqual(b2.run(s).result().x, (s.x + 1) * 7 * 7 + 1)

        with self.assertRaises(TypeError):
            a | 1
        with self.assertRaises(TypeError):
            b1 | 1

    def test_look_and_feel(self):
        class A(Runnable): pass
        class B(Runnable): pass

        b = A() | B()
        self.assertEqual(b.name, 'Branch')
        self.assertEqual(str(b), 'A | B')
        self.assertEqual(repr(b), 'Branch(components=(A(), B()))')
        self.assertEqual(tuple(b), b.components)
        self.assertIsInstance(b, Branch)
        self.assertIsInstance(b | b, Branch)

    def test_error_prop(self):
        class ErrorSilencer(Runnable):
            def next(self, state):
                return state
            def error(self, exc):
                return State(error=True)

        class Identity(Runnable):
            def next(self, state):
                return state

        branch = ErrorSilencer() | Identity()
        s1 = Present(exception=KeyError())
        s2 = branch.run(s1).result()

        self.assertEqual(s2.error, True)

    def test_stop(self):
        class Stoppable(Runnable):
            def init(self, state):
                self.stopped = False
            def next(self, state):
                return state
            def halt(self):
                self.stopped = True

        branch = Stoppable() | Stoppable()
        branch.run(State())
        branch.stop()

        self.assertTrue(all(c.stopped for c in branch.components))


class TestBranches(unittest.TestCase):

    def test_empty(self):
        with self.assertRaises(ValueError):
            Branches()

    def test_composition(self):
        class A(Runnable):
            def next(self, state):
                return state.updated(x=state.x + 1)
        class B(Runnable):
            def next(self, state):
                return state.updated(x=state.x * 7)

        a, b = A(), B()

        # single branch

        b1 = Branches(a)
        ss = States(State(x=1))
        res = b1.run(ss).result()

        self.assertEqual(b1.branches, (a,))
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].x, ss[0].x + 1)

        # two branches, explicit and implicit construction

        for b2 in [Branches(a, b), a & b]:
            ss = States(State(x=1), State(x=1))
            res = b2.run(ss).result()

            self.assertEqual(b2.branches, (a, b))
            self.assertEqual(len(res), 2)
            self.assertEqual(res[0].x, ss[0].x + 1)
            self.assertEqual(res[1].x, ss[1].x * 7)

        # appending a branch to branches

        b3 = b2 & a
        ss = States(*[State(x=1) for _ in range(3)])
        res = b3.run(ss).result()

        self.assertEqual(b3.branches, (a, b, a))
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0].x, ss[0].x + 1)
        self.assertEqual(res[1].x, ss[1].x * 7)
        self.assertEqual(res[2].x, ss[2].x + 1)

        # prepending a branch to branches
        b4 = b & b2
        ss = States(*[State(x=1) for _ in range(3)])
        res = b4.run(ss).result()

        self.assertEqual(b4.branches, (b, a, b))
        self.assertEqual(len(res), 3)
        self.assertEqual(res[0].x, ss[0].x * 7)
        self.assertEqual(res[1].x, ss[1].x + 1)
        self.assertEqual(res[2].x, ss[2].x * 7)

        # invalid type

        with self.assertRaises(TypeError):
            b & 1
        with self.assertRaises(TypeError):
            b1 & 1

    def test_look_and_feel(self):
        br = Runnable(), Runnable()
        pb = Branches(*br)
        self.assertEqual(pb.name, 'Branches')
        self.assertEqual(str(pb), '(Runnable) & (Runnable)')
        self.assertEqual(repr(pb), 'Branches(Runnable(), Runnable())')
        self.assertEqual(tuple(pb), br)

    def test_continuity(self):
        class Fast(Runnable):
            def next(self, state):
                time.sleep(0.1)
                return state.updated(x=state.x + 1)

        class Slow(Runnable):
            def next(self, state):
                time.sleep(0.2)
                return state.updated(x=state.x + 2)

        bs = Branches(Slow(), Fast(), Slow())
        ss = States(*[State(x=0) for _ in range(3)])
        res = bs.run(ss).result()
        self.assertEqual([s.x for s in res], [2, 1, 2])

    def test_parallel_independent_execution(self):
        class Component(Runnable):
            def __init__(self, runtime):
                super(Component, self).__init__()
                self.runtime = runtime
            def next(self, state):
                time.sleep(self.runtime)
                return state

        # make sure all branches really run in parallel
        n = 5
        bs = Branches(*[Component(1) for _ in range(n)])
        ss = States(*[State() for _ in range(n)])
        with tictoc() as tt:
            bs.run(ss).result()

        # total runtime has to be smaller that the sum of individual runtimes
        self.assertTrue(1 <= tt.dt <= 2)


class TestDup(unittest.TestCase):

    def test_basic(self):
        s = State(x=1)
        self.assertEqual(Dup(0).run(s).result(), States())
        self.assertEqual(Dup(1).run(s).result(), States(s))
        self.assertEqual(Dup(2).run(s).result(), States(s, s))

    def test_immutability(self):
        s = State(x=1)
        ss = Dup(1).run(s).result()

        s.x = 2
        self.assertEqual(ss[0].x, 1)


class TestRacingBranches(unittest.TestCase):

    def test_look_and_feel(self):
        br = Runnable(), Runnable()
        rb = RacingBranches(*br)
        self.assertEqual(rb.name, 'RacingBranches')
        self.assertEqual(str(rb), '(Runnable) !! (Runnable)')
        self.assertEqual(repr(rb), 'RacingBranches(Runnable(), Runnable())')
        self.assertEqual(tuple(rb), br)

    def test_stopped(self):
        class Fast(Runnable):
            def next(self, state, **runopts):
                time.sleep(0.1)
                return state.updated(x=state.x + 1)

        class Slow(Runnable):
            def init(self, state, **runopts):
                self.time_to_stop = threading.Event()

            def next(self, state, **runopts):
                self.time_to_stop.wait()
                return state.updated(x=state.x + 2)

            def halt(self):
                self.time_to_stop.set()

        # default case
        rb = RacingBranches(Slow(), Fast(), Slow())
        res = rb.run(State(x=0)).result()
        self.assertEqual([s.x for s in res], [2, 1, 2])

        # "endomorphic case"
        rb = RacingBranches(BlockingIdentity(), Slow(), Fast(), Slow())
        res = rb.run(State(x=0)).result()
        self.assertEqual([s.x for s in res], [0, 2, 1, 2])


class TestParallelBranches(unittest.TestCase):

    def test_look_and_feel(self):
        br = Runnable(), Runnable()
        pb = ParallelBranches(*br)
        self.assertEqual(pb.name, 'ParallelBranches')
        self.assertEqual(repr(pb), 'ParallelBranches(Runnable(), Runnable())')
        self.assertEqual(tuple(pb), br)

    def test_basic(self):
        class Fast(Runnable):
            def next(self, state):
                time.sleep(0.1)
                return state.updated(x=state.x + 1)

        class Slow(Runnable):
            def next(self, state):
                time.sleep(0.2)
                return state.updated(x=state.x + 2)

        # "endomorphic case"
        pb = ParallelBranches(Identity(), Slow(), Fast(), Slow())
        res = pb.run(State(x=0)).result()
        self.assertEqual([s.x for s in res], [0, 2, 1, 2])

        # default case
        pb = ParallelBranches(Slow(), Fast(), Slow())
        res = pb.run(State(x=0)).result()
        self.assertEqual([s.x for s in res], [2, 1, 2])

    def test_parallel_independent_execution(self):
        class Component(Runnable):
            def __init__(self, runtime):
                super(Component, self).__init__()
                self.runtime = runtime
            def next(self, state):
                time.sleep(self.runtime)
                return state

        # make sure all branches really run in parallel
        pb = ParallelBranches(
            Component(1), Component(1), Component(1), Component(1), Component(1))
        with tictoc() as tt:
            pb.run(State()).result()

        # total runtime has to be smaller that the sum of individual runtimes
        self.assertTrue(1 <= tt.dt <= 2)

    def test_mimo(self):
        """Parallel should support branches with States inputs"""

        inp = States(State(x=1), State(x=2))
        wrk = ParallelBranches(Identity(), Identity(), Identity())
        out = wrk.run(inp).result()

        self.assertEqual(len(out), 3)
        self.assertIsInstance(out, States)
        self.assertEqual(out.first, inp)


class TestArgMin(unittest.TestCase):

    def test_look_and_feel(self):
        fold = ArgMin(key=False)
        self.assertEqual(fold.name, 'ArgMin')
        self.assertEqual(str(fold), '[]>')
        self.assertEqual(repr(fold), "ArgMin(key=False)")

        fold = ArgMin(key=min)
        self.assertEqual(repr(fold), "ArgMin(key=<built-in function min>)")

    def test_default_fold(self):
        bqm = dimod.BinaryQuadraticModel({'a': 1}, {}, 0, dimod.SPIN)
        states = States(
            State.from_sample(min_sample(bqm), bqm),    # energy: -1
            State.from_sample(max_sample(bqm), bqm),    # energy: +1
        )
        best = ArgMin().run(states).result()
        self.assertEqual(best.samples.first.energy, -1)

    def test_custom_fold(self):
        bqm = dimod.BinaryQuadraticModel({'a': 1}, {}, 0, dimod.SPIN)
        states = States(
            State.from_sample(min_sample(bqm), bqm),    # energy: -1
            State.from_sample(max_sample(bqm), bqm),    # energy: +1
        )
        fold = ArgMin(key=lambda s: -s.samples.first.energy)
        best = fold.run(states).result()
        self.assertEqual(best.samples.first.energy, 1)


class TestTrackMin(unittest.TestCase):

    def test_look_and_feel(self):
        tracker = TrackMin(key=False, output=False, output_key='a', input_key='a')
        self.assertEqual(tracker.name, 'TrackMin')
        self.assertEqual(str(tracker), 'TrackMin')
        self.assertEqual(repr(tracker), "TrackMin(key=False, output=False, input_key='a', output_key='a')")

    def test_default_tracking(self):
        """Best seen state is kept (default: state with sampleset with the lowest energy)"""

        bqm = dimod.BinaryQuadraticModel({'a': 1}, {}, 0, dimod.SPIN)
        min_state = State.from_sample(min_sample(bqm), bqm)    # energy: -1
        max_state = State.from_sample(max_sample(bqm), bqm)    # energy: +1

        tracker = TrackMin()
        _ = tracker.run(max_state).result()
        self.assertEqual(tracker.best.samples.first.energy, +1)
        _ = tracker.run(min_state).result()
        self.assertEqual(tracker.best.samples.first.energy, -1)
        _ = tracker.run(max_state).result()
        self.assertEqual(tracker.best.samples.first.energy, -1)

    def test_custom_key(self):
        """Custom key function works, here best state has the highest energy."""

        bqm = dimod.BinaryQuadraticModel({'a': 1}, {}, 0, dimod.SPIN)
        states = States(
            State.from_sample(min_sample(bqm), bqm),    # energy: -1
            State.from_sample(max_sample(bqm), bqm),    # energy: +1
        )

        tracker = TrackMin(key=lambda s: -s.samples.first.energy)
        for state in states:
            tracker.run(state).result()
        self.assertEqual(tracker.best.samples.first.energy, +1)

    def test_output(self):
        """Best state is properly output, for a custom key."""

        state1 = State(a=1, en=1)
        state2 = State(a=2, en=0)

        tracker = TrackMin(key=lambda s: s.en, output=True, input_key='a', output_key='best')
        result1 = tracker.run(state1).result()
        self.assertEqual(result1.best, 1)
        result2 = tracker.run(state2).result()
        self.assertEqual(result2.best, 2)


class TestLoopUntilNoImprovement(unittest.TestCase):

    def test_max_iter(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        # iterate for `max_iter`
        it = LoopUntilNoImprovement(Inc(), max_iter=100, convergence=1000, key=lambda _: None)
        s = it.run(State(cnt=0)).result()
        self.assertEqual(s.cnt, 100)

        # `key` function not needed if `convergence` undefined
        it = LoopUntilNoImprovement(Inc(), max_iter=100, convergence=None)
        s = it.run(State(cnt=0)).result()
        self.assertEqual(s.cnt, 100)

        # `convergence` not needed for simple finite loop
        it = LoopUntilNoImprovement(Inc(), max_iter=100)
        s = it.run(State(cnt=0)).result()
        self.assertEqual(s.cnt, 100)

    def test_convergence(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        it = LoopUntilNoImprovement(Inc(), max_iter=1000, convergence=100, key=lambda _: None)
        s = it.run(State(cnt=0)).result()

        self.assertEqual(s.cnt, 100)

    def test_timeout(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        # timeout after exactly two runs
        with mock.patch('time.time', side_effect=itertools.count(), create=True):
            loop = LoopUntilNoImprovement(Inc(), max_time=2)
            state = loop.run(State(cnt=0)).result()
            self.assertEqual(state.cnt, 2)

        # timeout after the second run
        with mock.patch('time.time', side_effect=itertools.count(), create=True):
            loop = LoopUntilNoImprovement(Inc(), max_time=2.5)
            state = loop.run(State(cnt=0)).result()
            self.assertEqual(state.cnt, 3)

    def test_terminate_predicate(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        it = LoopUntilNoImprovement(Inc(),
                                    key=lambda state: state.cnt,
                                    terminate=lambda key: key >= 3)
        s = it.run(State(cnt=0)).result()

        self.assertEqual(s.cnt, 3)

    def test_energy_threshold_termination(self):
        class ExactSolver(Runnable):
            def next(self, state):
                return state.updated(
                    samples=dimod.ExactSolver().sample(state.problem))

        bqm = dimod.BinaryQuadraticModel({'a': 1}, {}, 0, dimod.SPIN)
        state = State.from_sample({'a': 1}, bqm)

        w = LoopUntilNoImprovement(ExactSolver(),
                                   key=operator.attrgetter('samples.first.energy'),
                                   terminate=partial(operator.ge, -1))
        s = w.run(state).result()
        self.assertEqual(s.samples.first.energy, -1)

        w = LoopUntilNoImprovement(ExactSolver(),
                                   key='samples.first.energy',
                                   terminate=partial(operator.ge, -1))
        s = w.run(state).result()
        self.assertEqual(s.samples.first.energy, -1)

        w = LoopUntilNoImprovement(ExactSolver(),
                                   terminate=partial(operator.ge, -1))
        s = w.run(state).result()
        self.assertEqual(s.samples.first.energy, -1)

    def test_finite_loop(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        it = Loop(Inc(), 10)
        s = it.run(State(cnt=0)).result()

        self.assertEqual(s.cnt, 10)

    def test_infinite_loop_stops_before_first_run(self):
        """An infinite loop can be preemptively stopped (before starting)."""

        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        loop = Loop(Inc())
        loop.stop()
        state = loop.run(State(cnt=0))

        self.assertEqual(state.result().cnt, 0)

    def test_infinite_loop_stops(self):
        """An infinite loop can be stopped after at least one iteration."""

        class Inc(Runnable):
            def __init__(self):
                super(Inc, self).__init__()
                self.first_run = threading.Event()

            def next(self, state):
                self.first_run.set()
                return state.updated(cnt=state.cnt + 1)

        loop = Loop(Inc())
        state = loop.run(State(cnt=0))

        # make sure loop body runnable is run at least once, then issue stop
        loop.runnable.first_run.wait(timeout=1)
        loop.stop()

        self.assertTrue(state.result().cnt >= 1)

    def test_infinite_loop_runs_after_stop(self):
        """An infinite loop can be started again after being stopped."""

        class Inc(Runnable):
            def __init__(self):
                super(Inc, self).__init__()
                self.first_run = threading.Event()

            def next(self, state):
                self.first_run.set()
                return state.updated(cnt=state.cnt + 1)

        loop = Loop(Inc())
        state1 = loop.run(State(cnt=0))

        # make sure loop body runnable is run at least once, then issue stop
        loop.runnable.first_run.wait(timeout=1)
        loop.stop()

        # check the state AND make sure next() finishes (see #111)
        self.assertTrue(state1.result().cnt >= 1)

        # reset our test event object
        loop.runnable.first_run.clear()

        # run loop again
        state2 = loop.run(State(cnt=0))

        # make sure loop body runnable is run at least once, then issue stop
        loop.runnable.first_run.wait(timeout=1)
        loop.stop()

        self.assertTrue(state2.result().cnt >= 1)

    def test_infinite_loop_over_interruptable_runnable(self):
        """Stop signal must propagate to child runnable."""

        class IntInc(Runnable):
            def __init__(self):
                super(IntInc, self).__init__()
                self.time_to_stop = threading.Event()
                self.first_run = threading.Event()

            def next(self, state):
                self.first_run.set()
                self.time_to_stop.wait()
                return state.updated(cnt=state.cnt + 1)

            def halt(self):
                self.time_to_stop.set()

        loop = Loop(IntInc())
        s = loop.run(State(cnt=0))

        # make sure loop body runnable is run at least once
        loop.runnable.first_run.wait()

        loop.stop()

        self.assertTrue(s.result().cnt >= 1)

    def test_dynamic_validation(self):
        class simo(traits.SIMO, Runnable):
            def next(self, state):
                return States(state, state)

        with self.assertRaises(StateDimensionalityError):
            LoopUntilNoImprovement(simo()).run(State()).result()

        with self.assertRaises(StateDimensionalityError):
            LoopUntilNoImprovement(simo()).run(States()).result()

    def test_end_of_stream_termination(self):
        class Inc(Runnable):
            def next(self, state, **kw):
                cnt = state.cnt + 1
                if cnt > state.maxcnt:
                    raise EndOfStream
                return state.updated(cnt=cnt)

        loop = LoopWhileNoImprovement(Inc())
        state = loop.run(State(cnt=0, maxcnt=3)).result()

        self.assertEqual(state.cnt, 3)


class TestLoopWhileNoImprovement(unittest.TestCase):

    def test_no_improvement_tries(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        loop = LoopWhileNoImprovement(Inc(), max_tries=10, key=lambda _: 0)
        state = loop.run(State(cnt=0)).result()

        self.assertEqual(len(loop.runnable.timers['dispatch.next']), 10)
        self.assertEqual(state.cnt, 1)

    def test_max_iter(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        loop = LoopWhileNoImprovement(Inc(), max_iter=5, max_tries=10, key=lambda _: 0)
        state = loop.run(State(cnt=0)).result()

        self.assertEqual(len(loop.runnable.timers['dispatch.next']), 5)
        self.assertEqual(state.cnt, 1)

    def test_terminate_predicate(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        it = LoopWhileNoImprovement(Inc(),
                                    key=lambda state: state.cnt,
                                    terminate=lambda key: key >= 3)
        s = it.run(State(cnt=0)).result()

        self.assertEqual(s.cnt, 3)

    def test_infinite_loop_stops(self):
        """An infinite loop can be stopped after 10 iterations."""

        class Countdown(Runnable):
            """Countdown runnable that sets a semaphore on reaching zero."""

            def __init__(self):
                super(Countdown, self).__init__()
                self.ring = threading.Event()

            def next(self, state):
                output = state.updated(cnt=state.cnt - 1)
                if output.cnt <= 0:
                    self.ring.set()
                return output

        countdown = Countdown()
        loop = LoopWhileNoImprovement(countdown)
        state = loop.run(State(cnt=10))

        # stop only AFTER countdown reaches zero (10 iterations)
        # timeout in case Countdown failed before setting the flag
        countdown.ring.wait(timeout=1)
        loop.stop()

        self.assertTrue(state.result().cnt <= 0)

    def test_runs_with_improvement(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        loop = LoopWhileNoImprovement(Inc(), max_tries=100, key=lambda state: -min(3, state.cnt))
        state = loop.run(State(cnt=0)).result()

        self.assertEqual(len(loop.runnable.timers['dispatch.next']), 103)
        self.assertEqual(state.cnt, 4)

    def test_dynamic_validation(self):
        class simo(traits.SIMO, Runnable):
            def next(self, state):
                return States(state, state)

        with self.assertRaises(StateDimensionalityError):
            LoopWhileNoImprovement(simo()).run(State()).result()

        with self.assertRaises(StateDimensionalityError):
            LoopWhileNoImprovement(simo()).run(States()).result()


class TestMap(unittest.TestCase):

    def test_isolated(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        states = States(State(cnt=1), State(cnt=2))
        result = Map(Inc()).run(states).result()

        self.assertEqual(result[0].cnt, states[0].cnt + 1)
        self.assertEqual(result[1].cnt, states[1].cnt + 1)

    def test_branch(self):
        class Inc(Runnable):
            def next(self, state):
                return state.updated(cnt=state.cnt + 1)

        states = States(State(cnt=1), State(cnt=2))
        branch = Map(Inc()) | ArgMin('cnt')
        result = branch.run(states).result()

        self.assertEqual(result.cnt, states[0].cnt + 1)

    def test_input_validation(self):
        with self.assertRaises(TypeError):
            Map(False)
        with self.assertRaises(TypeError):
            Map(lambda: None)
        with self.assertRaises(TypeError):
            Map(Runnable)
        self.assertIsInstance(Map(Runnable()), Runnable)


class TestReduce(unittest.TestCase):

    class Sum(traits.MISO, Runnable):
        def next(self, states):
            a, b = states
            return a.updated(val=a.val + b.val)

    def test_basic(self):
        states = States(State(val=1), State(val=2), State(val=3))
        result = Reduce(self.Sum()).run(states).result()

        self.assertIsInstance(result, State)
        self.assertEqual(result.val, 1+2+3)

    def test_initial_state(self):
        initial = State(val=10)
        states = States(State(val=1), State(val=2))
        result = Reduce(self.Sum(), initial_state=initial).run(states).result()

        self.assertEqual(result.val, 10+1+2)

    def test_unstructured_runnable(self):
        initial = State(val=10)
        states = States(State(val=2), State(val=3))

        multiply = Lambda(next=lambda self, s: s[0].updated(val=s[0].val * s[1].val))
        result = Reduce(multiply, initial_state=initial).run(states).result()

        self.assertEqual(result.val, 10*2*3)

    def test_input_validation(self):
        with self.assertRaises(TypeError):
            Reduce(False)
        with self.assertRaises(TypeError):
            Reduce(Runnable)
        self.assertIsInstance(Reduce(self.Sum()), Runnable)


class TestLambda(unittest.TestCase):

    def test_basic_runnable(self):
        runnable = Lambda(lambda _, s: s.updated(c=s.a * s.b))
        state = State(a=2, b=3)
        result = runnable.run(state).result()

        self.assertEqual(result.c, state.a * state.b)

    def test_error_and_init(self):
        runnable = Lambda(
            next=lambda self, state: state.updated(c=state.a * state.b),
            error=lambda self, exc: State(error=exc),
            init=lambda self, state: setattr(self, 'first', state.c)
        )

        # test init
        state = State(a=2, b=3, c=0)
        result = runnable.run(state).result()

        self.assertEqual(runnable.first, 0)
        self.assertEqual(result.c, state.a * state.b)

        # test error prop
        exc = ZeroDivisionError()
        result = runnable.run(Present(exception=exc)).result()

        self.assertEqual(result.error, exc)

    def test_map_lambda(self):
        states = States(State(cnt=1), State(cnt=2))
        result = Map(Lambda(lambda _, s: s.updated(cnt=s.cnt + 1))).run(states).result()

        self.assertEqual(result[0].cnt, states[0].cnt + 1)
        self.assertEqual(result[1].cnt, states[1].cnt + 1)

    def test_input_validation(self):
        with self.assertRaises(TypeError):
            Lambda(False)
        with self.assertRaises(TypeError):
            Lambda(lambda: None, False)
        with self.assertRaises(TypeError):
            Lambda(lambda: None, lambda: None, False)
        self.assertIsInstance(Lambda(lambda: None, lambda: None, lambda: None), Runnable)


class TestUnwind(unittest.TestCase):

    def test_basic(self):
        class Streamer(Runnable):
            def next(self, state, **runopts):
                if state.cnt <= 0:
                    raise EndOfStream
                return state.updated(cnt=state.cnt - 1)

        r = Unwind(Streamer())
        states = r.run(State(cnt=3)).result()

        # states should contain 3 states with cnt=3..0
        self.assertEqual(len(states), 3)
        for idx, state in enumerate(states):
            self.assertEqual(state.cnt, 2-idx)

    def test_dynamic_validation(self):
        class simo(traits.SIMO, Runnable):
            def next(self, state):
                return States(state, state)

        with self.assertRaises(StateDimensionalityError):
            Unwind(simo()).run(States()).result()


class TestIdentity(unittest.TestCase):

    def test_basic(self):
        ident = Identity()
        state = State(x=1, y='a', z=[1,2,3])

        inp = copy.deepcopy(state)
        out = ident.run(state).result()

        self.assertEqual(inp, out)
        self.assertFalse(out is inp)

    def test_input_type_invariant(self):
        inp1 = State(x=1)
        self.assertEqual(Identity().run(inp1).result(), inp1)

        inp2 = States(State(x=1), State(x=2))
        self.assertEqual(Identity().run(inp2).result(), inp2)


class TestBlockingIdentity(unittest.TestCase):

    def test_basic(self):
        ident = BlockingIdentity()
        state = State(x=1, y='a', z=[1,2,3])

        inp = copy.deepcopy(state)
        fut = ident.run(state)
        ident.stop()
        out = fut.result()

        self.assertEqual(inp, out)
        self.assertFalse(out is inp)

    def test_interruptable(self):
        ident = BlockingIdentity()
        state = State(x=1)
        out = ident.run(state)

        # ident block should not finish
        done, not_done = futures.wait({out}, timeout=0.1)
        self.assertEqual(len(done), 0)
        self.assertEqual(len(not_done), 1)

        # until we stop it
        ident.stop()

        done, not_done = futures.wait({out}, timeout=0.1)
        self.assertEqual(len(done), 1)
        self.assertEqual(len(not_done), 0)

        self.assertEqual(out.result().x, 1)

    def test_input_type_invariant(self):
        inp = State(x=1)
        ii = BlockingIdentity()
        fut = ii.run(inp)
        ii.stop()
        out = fut.result()
        self.assertEqual(out, inp)

        inp = States(State(x=1), State(x=2))
        ii = BlockingIdentity()
        fut = ii.run(inp)
        ii.stop()
        out = fut.result()
        self.assertEqual(out, inp)


class TestWait(unittest.TestCase):

    def test_basic(self):
        wait = Wait()
        state = State(x=1)
        out = wait.run(state)

        # wait block should not finish
        done, not_done = futures.wait({out}, timeout=0.1)
        self.assertEqual(len(done), 0)
        self.assertEqual(len(not_done), 1)

        # until we stop it
        wait.stop()

        done, not_done = futures.wait({out}, timeout=0.1)
        self.assertEqual(len(done), 1)
        self.assertEqual(len(not_done), 0)

        self.assertEqual(out.result().x, 1)

    def test_input_type_invariant(self):
        inp = State(x=1)
        w = Wait()
        r = w.run(inp)
        w.stop()
        self.assertEqual(r.result(), inp)

        inp = States(State(x=1), State(x=2))
        w = Wait()
        r = w.run(inp)
        w.stop()
        self.assertEqual(r.result(), inp)


class TestConst(unittest.TestCase):

    def test_nop(self):
        """For empty Const, cutput is a deepcopy of input, otherwise identical."""

        const = Const()
        inp = State()
        out = const.run(inp).result()

        self.assertEqual(inp, out)
        self.assertFalse(inp is out)

    def test_set_state_var(self):
        """State variable is properly set/reset/updated."""

        self.assertEqual(Const(x=1).run(State()).result().x, 1)
        self.assertEqual(Const(x=1).run(State(x=0)).result().x, 1)
        self.assertEqual(Const(x=None).run(State(x=0)).result().x, None)

    def test_set_states_var(self):
        """All States have variable set."""

        wrk = Const(x=1)
        inp = States(State(), State(x=0), State(x=None))
        exp = States(State(x=1), State(x=1), State(x=1))
        self.assertEqual(wrk.run(inp).result(), exp)

        wrk = Const(x=None)
        inp = States(State(), State(x=1))
        exp = States(State(x=None), State(x=None))
        self.assertEqual(wrk.run(inp).result(), exp)


class TestLog(unittest.TestCase):

    class Inc(Runnable):
        def next(self, state):
            return state.updated(x=state.x + 1)

    def test_log_output(self):
        with tempfile.TemporaryFile(mode='w+t') as fp:
            tag = 'test'
            log = Log(key=lambda state: state.x, extra=dict(tag=tag), outfile=fp)
            inp = State(x=1)
            out = log.run(inp).result()

            # get lines logged to file
            fp.seek(0, os.SEEK_SET)
            lines = fp.readlines()
            self.assertEqual(len(lines), 1)

            record = json.loads(lines[0])
            self.assertIn('time', record)
            self.assertIn('timestamp', record)
            self.assertEqual(record['data'], inp.x)
            self.assertEqual(record['tag'], tag)

    def test_memo_in_a_loop(self):
        inc = TestLog.Inc()
        log = Log(key=lambda state: state.x, memo=True)
        wrk = Loop(inc | log, max_iter=3)
        inp = State(x=0)
        out = wrk.run(inp).result()

        self.assertEqual(len(log.records), 3)
        self.assertEqual([r['data'] for r in log.records], [1, 2, 3])

    def test_shared_memo_list(self):
        # list as shared, thread-safe, log record store
        memo = []
        key = operator.attrgetter('x')
        log1 = Log(key=key, memo=memo, extra=dict(id=1))
        log2 = Log(key=key, memo=memo, extra=dict(id=2))
        log3 = Log(key=key, memo=memo, extra=dict(id=3))
        inc = TestLog.Inc()
        wrk = log1 | inc | Parallel(log2, inc | log3)
        inp = State(x=1)
        out = wrk.run(inp).result()

        self.assertEqual(len(memo), 3)
        self.assertEqual({r['id'] for r in memo}, {1, 2, 3})
        self.assertEqual({r['data'] for r in memo}, {1, 2, 3})

    @parameterized.expand([
        (queue.Queue(), ),
        (queue.LifoQueue(), ),
        (queue.SimpleQueue(), ),
        (multiprocessing.Queue(), ),
    ])
    def test_shared_memo_queue(self, memo):
        # queue as shared, thread-safe, log record store
        key = operator.attrgetter('x')
        log1 = Log(key=key, memo=memo, extra=dict(id=1))
        log2 = Log(key=key, memo=memo, extra=dict(id=2))
        log3 = Log(key=key, memo=memo, extra=dict(id=3))
        inc = TestLog.Inc()
        wrk = log1 | inc | Parallel(log2, inc | log3)
        inp = State(x=1)
        out = wrk.run(inp).result()

        res = []
        while True:
            try:
                res.append(memo.get(timeout=1))
            except queue.Empty:
                break

        self.assertEqual(len(res), 3)
        self.assertEqual({r['id'] for r in res}, {1, 2, 3})
        self.assertEqual({r['data'] for r in res}, {1, 2, 3})

    def test_input_validation(self):
        with self.assertRaises(TypeError):
            Log(key=1)

    def test_logging(self):
        log = Log(key=lambda state: state.x, loglevel=logging.INFO)
        inp = State(x=1)

        with self.assertLogs('hybrid.flow', level='INFO') as logmgr:
            _ = log.run(inp).result()
            self.assertEqual(len(logmgr.output), 1)
            self.assertIn('Log', logmgr.output[0])
            self.assertIn('data', logmgr.output[0])



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

import numpy
from parameterized import parameterized

import dimod
import dwave_networkx as dnx
from dwave.cloud.utils import utcnow

from hybrid.core import SampleSet
from hybrid.utils import (
    chimera_tiles, flip_energy_gains, select_localsearch_adversaries,
    hstack_samplesets, NumpyEncoder, OceanEncoder)


class TestEnergyFlipGainUtils(unittest.TestCase):
    # minimized when variables differ
    notall = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'ca': 1}, 0, dimod.SPIN)

    # minimized when B is different from A and C
    notb = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'ca': -1}, 0, dimod.SPIN)

    def test_pos(self):
        """Correctly orders positive gains."""

        # flipping C makes for the highest energy gain
        gains = flip_energy_gains(self.notall, {'a': 1, 'b': 1, 'c': -1})
        self.assertEqual(gains, [(4.0, 'c'), (0.0, 'b'), (0.0, 'a')])
        gains = flip_energy_gains(self.notall, {'a': -1, 'b': -1, 'c': 1})
        self.assertEqual(gains, [(4.0, 'c'), (0.0, 'b'), (0.0, 'a')])

        # flipping any is equally bad
        gains = flip_energy_gains(self.notb, {'a': 1, 'b': -1, 'c': 1})
        self.assertEqual(gains, [(4.0, 'c'), (4.0, 'b'), (4.0, 'a')])

    def test_neg(self):
        """Correctly orders negative gains."""

        # flipping any is equally good
        gains = flip_energy_gains(self.notall, {'a': 1, 'b': 1, 'c': 1})
        self.assertEqual(gains, [(-4.0, 'c'), (-4.0, 'b'), (-4.0, 'a')])

        # flipping B is the worst
        gains = flip_energy_gains(self.notb, {'a': 1, 'b': 1, 'c': 1})
        self.assertEqual(gains, [(0.0, 'c'), (0.0, 'a'), (-4.0, 'b')])

    def test_subset(self):
        """Flip energy is correctly calculated for a subset of variables."""

        gains = flip_energy_gains(self.notb, {'a': 1, 'b': 1, 'c': 1}, {'b'})
        self.assertEqual(gains, [(-4.0, 'b')])

        gains = flip_energy_gains(self.notb, {'a': 1, 'b': 1, 'c': 1}, ['c'])
        self.assertEqual(gains, [(0.0, 'c')])

        gains = flip_energy_gains(self.notb, {'a': 1, 'b': 1, 'c': 1}, 'ab')
        self.assertEqual(gains, [(0.0, 'a'), (-4.0, 'b')])

    def test_sample_as_list(self):
        """List samples (index bqms) are properly handled."""

        bqm = dimod.BQM.from_ising({}, {(0, 1): 1})
        sample = [1, 1]

        gains = flip_energy_gains(bqm, sample)
        self.assertEqual(gains, [(-2, 1), (-2, 0)])

    def test_heterogeneous_variable_label_types(self):
        variables = ['a', ('b', 1), frozenset((1,2,3))]
        bqm = dimod.BQM.from_ising(zip(variables, [1, 2, 1]), {})
        gains = flip_energy_gains(bqm, zip(variables, [1, 1, 1]))

        self.assertEqual(
            gains, list(zip((-2.0, -2.0, -4.0),
                            (variables[2], variables[0], variables[1]))))

    def test_localsearch_adversaries(self):
        """When var flip increases energy."""

        defvar = select_localsearch_adversaries(self.notall, {'a': 1, 'b': 1, 'c': -1})
        self.assertEqual(defvar, ['c', 'b', 'a'])

        allvar = select_localsearch_adversaries(self.notall, {'a': 1, 'b': 1, 'c': -1}, max_n=3, min_gain=-1)
        self.assertEqual(allvar, ['c', 'b', 'a'])

        subvar = select_localsearch_adversaries(self.notall, {'a': 1, 'b': 1, 'c': -1}, max_n=2, min_gain=-1)
        self.assertEqual(subvar, ['c', 'b'])

        subvar = select_localsearch_adversaries(self.notall, {'a': 1, 'b': 1, 'c': -1}, min_gain=1)
        self.assertEqual(subvar, ['c'])

        nonevar = select_localsearch_adversaries(self.notall, {'a': 1, 'b': 1, 'c': -1}, min_gain=10)
        self.assertEqual(nonevar, [])

    def test_localsearch_friends(self):
        """When var flip decreases energy."""

        allvar = select_localsearch_adversaries(self.notb, {'a': 1, 'b': 1, 'c': 1}, min_gain=None)
        self.assertEqual(allvar, ['c', 'a', 'b'])

        neutral = select_localsearch_adversaries(self.notb, {'a': 1, 'b': 1, 'c': 1}, min_gain=0.0)
        self.assertEqual(neutral, ['c', 'a'])

        nonevar = select_localsearch_adversaries(self.notb, {'a': 1, 'b': 1, 'c': 1}, min_gain=1.0)
        self.assertEqual(nonevar, [])

        friends = select_localsearch_adversaries(self.notb, {'a': 1, 'b': 1, 'c': 1}, min_gain=-10)
        self.assertEqual(friends, ['c', 'a', 'b'])


class TestChimeraTiles(unittest.TestCase):
    def test_single_target(self):
        bqm = dimod.BinaryQuadraticModel.from_qubo({edge: 1 for edge in dnx.chimera_graph(4).edges})

        tiles = chimera_tiles(bqm, 1, 1, 4)

        self.assertEqual(len(tiles), 16)  # we have the correct number of tiles
        self.assertEqual(set().union(*tiles.values()), bqm.variables)  # all of the nodes are present
        for embedding in tiles.values():
            self.assertEqual(set(chain[0] for chain in embedding.values()), set(range(1*1*4*2)))

    def test_even_divisor(self):
        bqm = dimod.BinaryQuadraticModel.from_qubo({edge: 1 for edge in dnx.chimera_graph(4).edges})

        tiles = chimera_tiles(bqm, 2, 2, 4)

        self.assertEqual(len(tiles), 4)  # we have the correct number of tiles
        self.assertEqual(set().union(*tiles.values()), bqm.variables)  # all of the nodes are present
        for embedding in tiles.values():
            self.assertEqual(set(chain[0] for chain in embedding.values()), set(range(2*2*4*2)))

    def test_uneven_divisor(self):
        si, sj, st = 3, 3, 4
        ti, tj, tt = 2, 2, 3
        bqm = dimod.BinaryQuadraticModel.from_qubo({edge: 1 for edge in dnx.chimera_graph(si, sj, st).edges})

        tiles = chimera_tiles(bqm, ti, tj, tt)

        self.assertEqual(len(tiles), 8)  # we have the correct number of tiles
        self.assertEqual(set().union(*tiles.values()), bqm.variables)  # all of the nodes are present
        for embedding in tiles.values():
            self.assertTrue(set(chain[0] for chain in embedding.values()).issubset(set(range(ti*tj*tt*2))))

    def test_string_labels(self):
        si, sj, st = 2, 2, 3
        ti, tj, tt = 1, 1, 4
        alpha = 'abcdefghijklmnopqrstuvwxyz'

        bqm = dimod.BinaryQuadraticModel.empty(dimod.BINARY)

        for u, v in reversed(list(dnx.chimera_graph(si, sj, st).edges)):
            bqm.add_interaction(alpha[u], alpha[v], 1)

        tiles = chimera_tiles(bqm, ti, tj, tt)

        self.assertEqual(len(tiles), 4)  # we have the correct number of tiles
        self.assertEqual(set().union(*tiles.values()), bqm.variables)  # all of the nodes are present
        for embedding in tiles.values():
            self.assertTrue(set(chain[0] for chain in embedding.values()).issubset(set(range(ti*tj*tt*2))))


class TestSampleSetUtils(unittest.TestCase):

    def test_hstack_empty(self):
        s1 = SampleSet.empty()
        s2 = SampleSet.empty()
        exp = SampleSet.empty()

        res = hstack_samplesets(s1, s2)
        self.assertEqual(res, exp)

    def test_hstack_identity(self):
        """hstack_samplesets is identity op on single sampleset's samples"""
        ss = dimod.SampleSet.from_samples({'a': 1}, vartype='BINARY', energy=0)
        hs = hstack_samplesets(ss)

        self.assertEqual(hs, ss)
        self.assertEqual(hs.record.sample.dtype, ss.record.sample.dtype)

    def test_hstack_with_empty(self):
        ss = dimod.SampleSet.from_samples({'a': 1}, vartype='BINARY', energy=0)
        emp = SampleSet.empty()

        self.assertEqual(hstack_samplesets(ss, emp), ss)
        self.assertEqual(hstack_samplesets(emp, ss), ss)

    def test_hstack_combine(self):
        s1 = dimod.SampleSet.from_samples({'a': 1}, vartype='BINARY', energy=0)
        s2 = dimod.SampleSet.from_samples({'b': 1}, vartype='BINARY', energy=0)
        exp = dimod.SampleSet.from_samples({'a': 1, 'b': 1}, vartype='BINARY', energy=0)

        self.assertEqual(hstack_samplesets(s1, s2), exp)
        self.assertEqual(hstack_samplesets(s2, s1), exp)

    def test_hstack_correct_dtype(self):
        """Output sampleset has the same dtype"""
        s1 = dimod.SampleSet.from_samples({'a': 1}, vartype='BINARY', energy=0)
        s2 = dimod.SampleSet.from_samples({'b': 1}, vartype='BINARY', energy=0)
        hs = hstack_samplesets(s1, s2)

        self.assertEqual(s1.record.sample.dtype, s2.record.sample.dtype)
        self.assertEqual(hs.record.sample.dtype, s1.record.sample.dtype)

    def test_hstack_clamp(self):
        s1 = dimod.SampleSet.from_samples([{'a': 1}, {'a': 0}], vartype='BINARY', energy=0)
        s2 = dimod.SampleSet.from_samples({'b': 1}, vartype='BINARY', energy=0)
        exp = dimod.SampleSet.from_samples({'a': 1, 'b': 1}, vartype='BINARY', energy=0)

        self.assertEqual(hstack_samplesets(s1, s2), exp)
        self.assertEqual(hstack_samplesets(s2, s1), exp)

    def test_hstack_update(self):
        s1 = dimod.SampleSet.from_samples({'a': 1}, vartype='BINARY', energy=0)
        s2 = dimod.SampleSet.from_samples({'a': 0}, vartype='BINARY', energy=0)

        self.assertEqual(hstack_samplesets(s1, s2), s2)
        self.assertEqual(hstack_samplesets(s2, s1), s1)

    def test_hstack_multisample_multivar(self):
        ab = dimod.SampleSet.from_samples(
            [{'a': 0, 'b': 1}, {'a': 1, 'b': 0}], vartype='BINARY', energy=0)
        bc = dimod.SampleSet.from_samples(
            [{'b': 1, 'c': 0}, {'b': 1, 'c': 1}], vartype='BINARY', energy=0)
        exp = dimod.SampleSet.from_samples(
            [{'a': 0, 'b': 1, 'c': 0}, {'a': 1, 'b': 1, 'c': 1}], vartype='BINARY', energy=0)

        self.assertEqual(hstack_samplesets(ab, bc), exp)

    def test_hstack_from_bqm(self):
        bqm = dimod.BQM.from_ising({'a': 1}, {})
        ss = dimod.SampleSet.from_samples({'a': 0}, vartype='BINARY', energy=0)

        res = hstack_samplesets(ss, bqm=bqm)
        self.assertEqual(res.vartype, dimod.SPIN)
        numpy.testing.assert_array_equal(res.record.energy, numpy.array([-1]))


class TestNumpyJSONEncoder(unittest.TestCase):

    @parameterized.expand([
        (numpy.bool_(1), True),
        (numpy.byte(1), 1), (numpy.int8(1), 1),
        (numpy.ubyte(1), 1), (numpy.uint8(1), 1),
        (numpy.short(1), 1), (numpy.int16(1), 1),
        (numpy.ushort(1), 1), (numpy.uint16(1), 1),
        (numpy.intc(1), 1), (numpy.int32(1), 1),
        (numpy.uintc(1), 1), (numpy.uint32(1), 1),
        (numpy.int_(1), 1), (numpy.int32(1), 1),
        (numpy.uint(1), 1), (numpy.uint32(1), 1),
        (numpy.longlong(1), 1), (numpy.int64(1), 1),
        (numpy.ulonglong(1), 1), (numpy.uint64(1), 1),
        (numpy.half(1.0), 1.0), (numpy.float16(1.0), 1.0),
        (numpy.single(1.0), 1.0), (numpy.float32(1.0), 1.0),
        (numpy.double(1.0), 1.0), (numpy.float64(1.0), 1.0),
        (numpy.longdouble(1.0), 1.0)
    ] + ([
        (numpy.float128(1.0), 1.0)      # unavailable on windows
    ] if hasattr(numpy, 'float128') else [
    ]))
    def test_numpy_primary_type_encode(self, np_val, py_val):
        self.assertEqual(
            json.dumps(py_val),
            json.dumps(np_val, cls=NumpyEncoder)
        )

    @parameterized.expand([
        (numpy.array([1, 2, 3], dtype=int), [1, 2, 3]),
        (numpy.array([[1], [2], [3]], dtype=float), [[1.0], [2.0], [3.0]]),
        (numpy.zeros((2, 2), dtype=bool), [[False, False], [False, False]]),
        (numpy.array([('Rex', 9, 81.0), ('Fido', 3, 27.0)],
                     dtype=[('name', 'U10'), ('age', 'i4'), ('weight', 'f4')]),
         [['Rex', 9, 81.0], ['Fido', 3, 27.0]]),
        (numpy.rec.array([(1, 2., 'Hello'), (2, 3., "World")],
                         dtype=[('foo', 'i4'), ('bar', 'f4'), ('baz', 'U10')]),
         [[1, 2.0, "Hello"], [2, 3.0, "World"]])
    ])
    def test_numpy_array_encode(self, np_val, py_val):
        self.assertEqual(
            json.dumps(py_val),
            json.dumps(np_val, cls=NumpyEncoder)
        )


class TestOceanJSONEncoder(unittest.TestCase):

    def test_datetime(self):
        dt = utcnow()
        self.assertEqual(
            json.dumps(dt, cls=OceanEncoder),
            json.dumps(dt.isoformat())
        )

    def test_bqm_encode(self):
        bqm = dimod.BQM.from_ising({'a': 1}, {'bc': 1})
        np_vec = {
            "linear": [1.0, 0.0, 0.0],
            "quadratic": [[1], [2], [1.0]],
            "offset": 0.0,
            "labels": ["a", "b", "c"]
        }
        self.assertEqual(
            json.dumps(bqm, cls=OceanEncoder),
            json.dumps(np_vec)
        )

    def test_sampleset_encode(self):
        ss = dimod.SampleSet.from_samples([1, 0, 1], 'BINARY', 0)
        self.assertEqual(
            json.dumps(ss, cls=OceanEncoder),
            json.dumps(ss.to_serializable())
        )



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import string
import itertools
import unittest
import collections
from functools import partial

import dimod
import networkx as nx
import dwave_networkx as dnx

from hybrid.decomposers import (
    EnergyImpactDecomposer, RandomSubproblemDecomposer,
    RandomConstraintDecomposer, RoofDualityDecomposer, ComponentDecomposer,
    SublatticeDecomposer, make_origin_embeddings,
    _make_cubic_lattice, _make_kings_lattice,
    _zephyr_to_chimeralike, _chimeralike_to_zephyr)
from hybrid.core import State
from hybrid.utils import min_sample, random_sample
from hybrid.exceptions import EndOfStream

from dwave.system.testing import MockDWaveSampler
from dwave.embedding import verify_embedding

class TestComponentDecomposer(unittest.TestCase):
    bqm = dimod.BinaryQuadraticModel({'a': 2, 'b': -1, 'd': 1}, {'bc': 1, 'cd': 1}, 0, dimod.SPIN)
    state = State.from_sample(random_sample(bqm), bqm)

    def test_default(self):
        decomposer = ComponentDecomposer()
        
        state1 = decomposer.next(self.state)
        self.assertIn(dict(state1.subproblem.linear), ({'a': 2}, {'b': -1, 'c': 0, 'd': 1}))

        state2 = decomposer.next(state1)
        self.assertIn(dict(state2.subproblem.linear), ({'a': 2}, {'b': -1, 'c': 0, 'd': 1}))
        self.assertNotEqual(dict(state1.subproblem.linear), dict(state2.subproblem.linear))

        state3 = decomposer.next(state2)  # silent_rewind=True, so rewind w/o raising an assertion
        self.assertIn(dict(state3.subproblem.linear), ({'a': 2}, {'b': -1, 'c': 0, 'd': 1}))

    def test_sort(self):
        decomposer = ComponentDecomposer(key=len)
        
        state1 = decomposer.next(self.state)
        self.assertDictEqual(dict(state1.subproblem.linear), {'b': -1, 'c': 0, 'd': 1})

    def test_sort_reverse(self):
        decomposer = ComponentDecomposer(key=len, reverse=False)
        
        state1 = decomposer.next(self.state)
        self.assertDictEqual(dict(state1.subproblem.linear), {'a': 2})

    def test_no_silent_rewind(self):
        decomposer = ComponentDecomposer(silent_rewind=False)
        state1 = decomposer.next(self.state)
        state2 = decomposer.next(state1)

        with self.assertRaises(EndOfStream):
            state3 = decomposer.next(state2)

    def test_key_func(self):
        def sum_linear_biases(component):
            total = 0
            for v in component:
                total += self.bqm.get_linear(v)
            return total
    
        decomposer = ComponentDecomposer(key=sum_linear_biases)

        state1 = decomposer.next(self.state)
        self.assertDictEqual(dict(state1.subproblem.linear), {'a': 2})

    def test_no_rolling(self):
        decomposer = ComponentDecomposer(rolling=False, key=len)

        state1 = decomposer.next(self.state)
        self.assertDictEqual(dict(state1.subproblem.linear), {'b': -1, 'c': 0, 'd': 1})

        state2 = decomposer.next(state1)
        self.assertDictEqual(dict(state2.subproblem.linear), {'b': -1, 'c': 0, 'd': 1})

    def test_one_component(self):
        bqm = dimod.BinaryQuadraticModel({'a': 1, 'b': -1}, {'ab': 1}, 0, dimod.SPIN)
        state = State.from_sample(random_sample(bqm), bqm)
        
        decomposer = ComponentDecomposer()
        state1 = decomposer.next(state)

        self.assertEqual(bqm, state1.subproblem)

    def test_empty(self):
        bqm = dimod.BinaryQuadraticModel(dimod.SPIN)
        state = State.from_sample(random_sample(bqm), bqm)

        decomposer = ComponentDecomposer()
        state1 = decomposer.next(state)

        self.assertEqual(bqm, state1.subproblem)

class TestEnergyImpactDecomposer(unittest.TestCase):
    # minimized when not all vars are equal
    notall = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'ca': 1}, 0, dimod.SPIN)

    def test_one_var(self):
        """First-variable selection works."""

        state = State.from_sample({'a': 1, 'b': 1, 'c': -1}, self.notall)
        eid = EnergyImpactDecomposer(size=1, min_gain=0)
        nextstate = eid.next(state)
        self.assertDictEqual(dict(nextstate.subproblem.linear), {'c': 2})
        self.assertDictEqual(dict(nextstate.subproblem.quadratic), {})

    def test_multi_vars(self):
        """Multiple variables subproblem selection works, without gain limit."""

        state = State.from_sample({'a': 1, 'b': 1, 'c': -1}, self.notall)
        eid = EnergyImpactDecomposer(size=3, min_gain=None)
        nextstate = eid.next(state)
        self.assertDictEqual(dict(nextstate.subproblem.adj), dict(self.notall.adj))

    def test_adaptive_vars(self):
        """Multiple variables subproblem selection works, with gain limit."""

        state = State.from_sample({'a': 1, 'b': 1, 'c': -1}, self.notall)
        eid = EnergyImpactDecomposer(size=3, min_gain=2.0)
        nextstate = eid.next(state)
        self.assertDictEqual(dict(nextstate.subproblem.linear), {'c': 2})
        self.assertDictEqual(dict(nextstate.subproblem.quadratic), {})

    def test_no_vars(self):
        """Failure due to no sub vars available."""

        state = State.from_sample({'a': 1, 'b': 1, 'c': -1}, self.notall)
        eid = EnergyImpactDecomposer(size=3, min_gain=5.0)
        nextstate = eid.next(state)
        self.assertEqual(len(nextstate.subproblem), 0)

    def test_rolling_subproblem(self):
        """Selected number of (non-overlapping) subproblems are unrolled from the input problem."""

        # 10 variables, 0 to 9 when ordered by energy increase on flip
        bqm = dimod.BinaryQuadraticModel({i: i for i in range(10)}, {}, 0.0, 'SPIN')
        sample = {i: 1 for i in range(10)}

        # exactly 5 single-variable problems should be produced
        state = State.from_sample(sample, bqm)
        eid = EnergyImpactDecomposer(size=1, rolling=True, rolling_history=0.5, silent_rewind=False)
        states = list(iter(partial(eid.next, state=state), None))

        self.assertEqual(len(states), 5)
        for idx, state in enumerate(states):
            self.assertEqual(state.subproblem.linear[idx], idx)

    def test_rolling_subproblem_larger_than_rolling_history(self):
        """In case rolling history too small, just one problem is unrolled."""

        # 10 variables, 0 to 9 when ordered by energy increase on flip
        bqm = dimod.BinaryQuadraticModel({i: i for i in range(10)}, {}, 0.0, 'SPIN')
        sample = {i: 1 for i in range(10)}

        # exactly 1 five-variable problems should be produced
        state = State.from_sample(sample, bqm)
        eid = EnergyImpactDecomposer(size=5, rolling=True, rolling_history=0.3, silent_rewind=False)
        states = list(iter(partial(eid.next, state=state), None))

        self.assertEqual(len(states), 1)
        self.assertEqual(len(states[0].subproblem), 5)
        self.assertEqual(list(dict(states[0].subproblem.linear).values()), list(range(0,5)))

        # works even for subproblems as large as the input problem
        eid = EnergyImpactDecomposer(size=len(bqm), rolling=True, rolling_history=0.3, silent_rewind=False)
        states = list(iter(partial(eid.next, state=state), None))

        self.assertEqual(len(states), 1)
        self.assertEqual(len(states[0].subproblem), 10)
        self.assertEqual(list(dict(states[0].subproblem.linear).values()), list(range(0,10)))

        # but adapt to problem size if subproblem larger than problem
        eid = EnergyImpactDecomposer(size=11, rolling=True, rolling_history=0.3, silent_rewind=False)
        states = list(iter(partial(eid.next, state=state), None))

        self.assertEqual(len(states), 1)
        self.assertEqual(len(states[0].subproblem), 10)
        self.assertEqual(list(dict(states[0].subproblem.linear).values()), list(range(0,10)))

    def test_energy_traverse(self):
        eid = EnergyImpactDecomposer(size=1, traversal='energy')
        priority = collections.OrderedDict((v,v) for v in range(1,4))
        var = eid.traverse(bqm=None, sample=None, ordered_priority=priority, visited=[2,1], size=2)
        self.assertEqual(var, [3])

    def test_bfs_traverse_connected(self):
        eid = EnergyImpactDecomposer(size=None, traversal='bfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('bdac', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set(), size=3)
        # start from 'b', pick 2 more neighbors
        self.assertEqual(var, set('abc'))

    def test_bfs_traverse_connected_too_small(self):
        eid = EnergyImpactDecomposer(size=None, traversal='bfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('bdac', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set(), size=5)
        # start from 'b', try to pick more then there is (get complete graph back)
        self.assertEqual(var, set('abcd'))

    def test_bfs_traverse_connected_some_visited(self):
        eid = EnergyImpactDecomposer(size=None, traversal='bfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1, 'bd': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('bdac', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set('b'), size=3)
        # start with 'b' visited, so use 'd' as root and pick 2 more neighbors
        self.assertEqual(var, set('dac'))

    def test_bfs_traverse_disconnected_some_visited(self):
        eid = EnergyImpactDecomposer(size=None, traversal='bfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1, 'ef': 1, 'fg': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('abcdefg', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set('abc'), size=3)
        # pick 'd' from first graph component, and 'ef' from the second
        self.assertEqual(var, set('def'))

    def test_bfs_on_sequential_eid_calls_over_disconnected_problem_graph(self):
        # problem graph has two components, each one is 4-node cycle graph
        edges = {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1,
                 'ef': 1, 'fg': 1, 'gh': 1, 'he': 1}
        biases = dict(zip(string.ascii_letters, range(8, 0, -1)))
        bqm = dimod.BinaryQuadraticModel(biases, edges, 0.0, 'SPIN')
        sample = {i: -1 for i in bqm.variables}

        state = State.from_sample(sample, bqm)
        eid = EnergyImpactDecomposer(size=3, traversal='bfs', rolling=True,
                                     rolling_history=1.0, silent_rewind=False)
        states = list(iter(partial(eid.next, state=state), None))

        # energy impact list is: [a..h], so of the 3 subproblems generated,
        # the middle one is disconnected with one var from first group and two
        # variables from the second

        # `c` has higher energy, but it's not connected to `a`, so `d` is picked
        self.assertEqual(set(states[0].subproblem.variables), set('abd'))

        # `c` is picked from the first component, and the seed for the next
        # subproblem is `e`. however, the order of `e`'s neighbors is not defined,
        # so if we need to pick just one, it could be `f` or `h`
        # (note: PFS has a defined order of neighbors)
        self.assertTrue(set('cefh').difference(states[1].subproblem.variables).issubset('fh'))
        self.assertEqual(len(states[1].subproblem.variables), 3)

        # the second component is exhausted in search for 3 variable subproblem
        third = set(states[2].subproblem.variables)
        self.assertTrue(third == set('gh') or third == set('gf'))

    def test_nx_pfs_edgecases(self):
        pfs = EnergyImpactDecomposer._pfs_nodes
        graph = nx.Graph({'a': 'b', 'b': 'c', 'c': 'a'})
        priority = dict(zip('abc', itertools.count())).get

        self.assertEqual(set(pfs(graph, 'a', 0, priority)), set())
        self.assertEqual(set(pfs(graph, 'a', 1, priority)), set('a'))
        self.assertEqual(set(pfs(graph, 'a', 3, priority)), set('abc'))
        self.assertEqual(set(pfs(graph, 'a', 4, priority)), set('abc'))

    def test_nx_pfs_priority_respected(self):
        pfs = EnergyImpactDecomposer._pfs_nodes
        graph = nx.Graph({'a': 'b', 'b': 'c', 'c': 'a'})

        priority = dict(zip('abc', [1, 2, 3])).get
        self.assertEqual(set(pfs(graph, 'a', 2, priority)), set('ac'))
        self.assertEqual(set(pfs(graph, 'b', 2, priority)), set('bc'))
        self.assertEqual(set(pfs(graph, 'c', 2, priority)), set('cb'))

    def test_nx_pfs_deep_search(self):
        pfs = EnergyImpactDecomposer._pfs_nodes

        # two K4 graphs connected with one edge
        graph = nx.complete_graph(4)
        graph.add_edges_from(nx.complete_graph(range(4,8)).edges)
        graph.add_edge(1, 4)

        # use node index for weight/priority
        priority = lambda node: node

        # make sure once the second cluster is within reach, we deplete it
        self.assertEqual(set(pfs(graph, 3, 2, priority)), set([2, 3]))
        self.assertEqual(set(pfs(graph, 3, 3, priority)), set([1, 2, 3]))
        self.assertEqual(set(pfs(graph, 3, 4, priority)), set([1, 2, 3, 4]))
        self.assertEqual(set(pfs(graph, 3, 7, priority)), set([1, 2, 3, 4, 5, 6, 7]))
        self.assertEqual(set(pfs(graph, 3, 8, priority)), set([0, 1, 2, 3, 4, 5, 6, 7]))

    def test_pfs_traverse_connected(self):
        eid = EnergyImpactDecomposer(size=None, traversal='pfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('bdac', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set(), size=3)
        # start from 'b', pick 2 connected variables, traversed in order of best energy (priority)
        self.assertEqual(var, set('bad'))

    def test_pfs_traverse_connected_too_small(self):
        eid = EnergyImpactDecomposer(size=None, traversal='pfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('bdac', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set(), size=5)
        # start from 'b', try to pick more then there is (get complete graph back)
        self.assertEqual(var, set('abcd'))

    def test_pfs_traverse_connected_some_visited(self):
        eid = EnergyImpactDecomposer(size=None, traversal='pfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1, 'bd': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('bdac', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set('b'), size=3)
        # start with 'b' visited, so use 'd' as root and pick 2 more neighbors
        self.assertEqual(var, set('dac'))

    def test_pfs_traverse_disconnected_some_visited(self):
        eid = EnergyImpactDecomposer(size=None, traversal='pfs')
        bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1, 'ef': 1, 'fg': 1}, 0.0, dimod.SPIN)
        priority = collections.OrderedDict(zip('abcdefg', itertools.count(0, -1)))
        var = eid.traverse(bqm=bqm, sample=None, ordered_priority=priority, visited=set('abc'), size=3)
        # pick 'd' from first graph component, and 'ef' from the second
        self.assertEqual(var, set('def'))

    def test_pfs_on_sequential_eid_calls_over_disconnected_problem_graph(self):
        # problem graph has two components, each one is 4-node cycle graph
        edges = {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1,
                 'ef': 1, 'fg': 1, 'gh': 1, 'he': 1}
        biases = dict(zip(string.ascii_letters, range(8, 0, -1)))
        bqm = dimod.BinaryQuadraticModel(biases, edges, 0.0, 'SPIN')
        sample = {i: -1 for i in bqm.variables}

        state = State.from_sample(sample, bqm)
        eid = EnergyImpactDecomposer(size=3, traversal='pfs', rolling=True,
                                     rolling_history=1.0, silent_rewind=False)
        states = list(iter(partial(eid.next, state=state), None))

        # energy impact list is: [a..h], so of the 3 subproblems generated,
        # the middle one is disconnected with one var from first group and two
        # variables from the second

        # pfs is seeded with `a` and connected nodes in order of energy are picked
        self.assertEqual(set(states[0].subproblem.variables), set('abc'))

        # `d` left from the first component, and the seed for the next
        # subproblem is the next highest in energy `e`.
        # unlike in bfs, the order of `e`'s neighbors is well defined
        self.assertEqual(set(states[1].subproblem.variables), set('def'))

        # the second component is exhausted in search for 3 variable subproblem
        self.assertEqual(set(states[2].subproblem.variables), set('gh'))

    def test_pfs_on_impactful_far_subproblem(self):
        # problem graph has two components, each one is 4-node cycle graph
        # variable flip energy gains order variables: a, h..b
        edges = {'ab': 1, 'bc': 1, 'cd': 1, 'da': 1,
                 'ef': 1, 'fg': 1, 'gh': 1, 'he': 1,
                 'de': 0}
        biases = dict(zip(string.ascii_letters, range(8)))
        biases['a'] += 10
        bqm = dimod.BinaryQuadraticModel(biases, edges, 0.0, 'SPIN')
        sample = {i: -1 for i in bqm.variables}

        state = State.from_sample(sample, bqm)
        eid = EnergyImpactDecomposer(size=5, traversal='pfs')
        result = eid.run(state).result()

        # move towards second cluster and pick the highest energy variables from there
        self.assertEqual(set(result.subproblem.variables), set('adehg'))


class TestRandomSubproblemDecomposer(unittest.TestCase):
    bqm = dimod.BinaryQuadraticModel({}, {'ab': 1, 'bc': 1}, 0, dimod.SPIN)
    state = State.from_sample(min_sample(bqm), bqm)

    def test_simple(self):
        runnable = RandomSubproblemDecomposer(size=1)
        state = self.state
        for _ in range(10):
            state = runnable.next(state)
            self.assertEqual(len(state.subproblem.variables), 1)
            self.assertIn(next(iter(state.subproblem.variables)), self.bqm.variables)

    def test_look_and_feel(self):
        self.assertEqual(repr(RandomSubproblemDecomposer(7)), 'RandomSubproblemDecomposer(size=7)')


class TestSublatticeDecomposer(unittest.TestCase):
    """State fields bqm and origin_embeddings alongside optional state fields 
    geometric_offset, problem_dims, exclude_dims and origin_embedding_index
    have non-trivial dependencies that are not fully tested. Standard 
    operational modes are tested for square lattice models.
    """

    def test_trivial_subproblem(self):
        """Check single-variable no-coupler subproblem creation.

        A 2by2 square lattice (a square), with a single variable subsolver,
        is the simplest non-trivial test.
        """

        problem_dims = (2, 2)
        # Vertical edges
        edgelist = [((i, j), (i+1, j))
                    for i in range(problem_dims[0]-1)
                    for j in range(problem_dims[1])]
        # Horizontal edges
        edgelist += [((i, j), (i, j+1))
                     for i in range(problem_dims[0])
                     for j in range(problem_dims[1]-1)]
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {edge: 1 for edge in edgelist})
        origin_embeddings = [{(0, 0): [0]}]
        state = State.from_sample(min_sample(bqm), bqm,
                                  origin_embeddings=origin_embeddings,
                                  problem_dims=problem_dims)

        # Creates subproblems located randomly at (0, 0), (0, 1), (1, 0) and
        # (1, 1).
        runnable = SublatticeDecomposer()
        for _ in range(10):
            state = runnable.next(state)
            self.assertEqual(len(state.subproblem.variables), 1)
            self.assertIn(next(iter(state.subproblem.variables)), bqm.variables)
            self.assertEqual(len(state.embedding), 1)
            self.assertIn(next(iter(state.embedding.keys())), bqm.variables)

    def test_nontrivial_subproblem(self):
        """Check multi-variable multi-coupler subproblem creation.

        A 3by3 square lattice, with a 2x2 lattice subsolver, is the simplest 
        non-trivial test with edges in the decomposed problem.
        """
        problem_dims = (3, 3)
        #Vertical edges
        edgelist = [((i, j), (i+1, j))
                    for i in range(problem_dims[0]-1)
                    for j in range(problem_dims[1])]
        #Horizontal edges
        edgelist += [((i, j), (i, j+1))
                     for i in range(problem_dims[0])
                     for j in range(problem_dims[1]-1)]
        bqm = dimod.BinaryQuadraticModel.from_ising(
            {}, {edge: 1 for edge in edgelist})
        origin_embeddings = [{(i, j): None for i in range(2) for j in range(2)}]
        state = State.from_sample(min_sample(bqm), bqm,
                                  origin_embeddings=origin_embeddings,
                                  problem_dims=problem_dims)

        # Creates one of 3x3 different subsubproblems, some of which are
        # disconnected, and some connected 
        # Run multiple times to prevent coincidental agreement
        runnable = SublatticeDecomposer()
        for _ in range(10):
            state = runnable.next(state)
            self.assertEqual(len(state.subproblem.variables), 4)
            self.assertIn(next(iter(state.subproblem.variables)), bqm.variables)
            self.assertEqual(len(state.embedding), 4)
            self.assertIn(next(iter(state.embedding.keys())), bqm.variables)

    def test_geometric_offset(self):
        """Check exclude_dims parameter at maximum origin_embedding_index.
        
        If geometric offset is zero, the ``embedding`` field should match the 
        ``origin_embedding``.
        """
        problem_dims = (3, 3)
        # Vertical edges
        edgelist = [((i, j), (i+1, j))
                    for i in range(problem_dims[0]-1)
                    for j in range(problem_dims[1])]
        # Horizontal edges
        edgelist += [((i, j), (i, j+1))
                     for i in range(problem_dims[0])
                     for j in range(problem_dims[1]-1)]
        bqm = dimod.BinaryQuadraticModel.from_ising(
            {}, {edge: 1 for edge in edgelist})
        origin_embeddings = [{(i, j): None for i in range(2) for j in range(2)}]
        geometric_offset = (0,0)
        oei = len(origin_embeddings) - 1
        state = State.from_sample(min_sample(bqm), bqm,
                                  origin_embeddings=origin_embeddings,
                                  problem_dims=problem_dims,
                                  geometric_offset=geometric_offset,
                                  origin_embedding_index=oei)

        # Creates a 2x2 subproblem at the origin for final embedding
        # Run multiple times to prevent coincidental agreement
        runnable = SublatticeDecomposer()
        for _ in range(10):
            state = runnable.next(state)
            self.assertEqual(len(state.subproblem.variables), 4)
            self.assertIn(next(iter(state.subproblem.variables)), bqm.variables)
            self.assertEqual(len(state.embedding), 4)
            self.assertIn(next(iter(state.embedding.keys())), bqm.variables)
            self.assertIn(next(iter(state.embedding)), state.origin_embeddings[oei])
            
    def test_exclude_dimensions(self):
        """Check exclude_dims parameter at origin_embedding_index=0.

        If all dimensions are excluded, ``embedding`` field should match the 
        ``origin_embedding``.
        """
        problem_dims = (3, 3)
        # Vertical edges
        edgelist = [((i, j), (i+1, j))
                    for i in range(problem_dims[0]-1)
                    for j in range(problem_dims[1])]
        # Horizontal edges
        edgelist += [((i, j), (i, j+1))
                     for i in range(problem_dims[0])
                     for j in range(problem_dims[1]-1)]
        bqm = dimod.BinaryQuadraticModel.from_ising(
            {}, {edge: 1 for edge in edgelist})
        origin_embeddings = [{(i, j): None for i in range(2) for j in range(2)}]
        exclude_dims = list(range(len(problem_dims)))
        oei = 0
        state = State.from_sample(min_sample(bqm), bqm,
                                  origin_embeddings=origin_embeddings,
                                  problem_dims=problem_dims,
                                  exclude_dims=exclude_dims,
                                  origin_embedding_index=oei)

        # Creates a 2x2 subproblem at the origin for 0th embedding
        # Run multiple times to prevent coincidental agreement
        runnable = SublatticeDecomposer()
        for _ in range(10):
            state = runnable.next(state)
            self.assertEqual(len(state.subproblem.variables), 4)
            self.assertIn(next(iter(state.subproblem.variables)), bqm.variables)
            self.assertEqual(len(state.embedding), 4)
            self.assertIn(next(iter(state.embedding.keys())), bqm.variables)
            self.assertIn(next(iter(state.embedding)), state.origin_embeddings[oei])

class TestConstraintDecomposer(unittest.TestCase):
    def test_typical_construction(self):
        bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)

        variables = list('abcdefg')
        constraints = []
        for triplet in itertools.combinations(variables, 3):
            for u, v in itertools.combinations(triplet, 2):
                bqm.add_interaction(u, v, -1)
            constraints.append(triplet)

        rcd = RandomConstraintDecomposer(3, constraints)
        rcd.init(state=State.from_sample(min_sample(bqm), bqm))

        # check that the graph is complete
        G = rcd.constraint_graph
        for i in range(len(constraints)):
            self.assertIn(i, G.nodes)

    def test_next(self):
        bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)

        variables = list('abcdefg')
        constraints = []
        for triplet in itertools.combinations(variables, 3):
            for u, v in itertools.combinations(triplet, 2):
                bqm.add_interaction(u, v, -1)
            constraints.append(triplet)

        rcd = RandomConstraintDecomposer(3, constraints)

        state = State.from_sample(min_sample(bqm), bqm)

        newstate = rcd.run(state).result()

        self.assertIn('subproblem', newstate)
        self.assertTrue(len(newstate.subproblem) <= 3)  # correct size

    def test_next_on_different_sized_constraints(self):
        bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)

        variables = list('abcdefg')
        fixed_variables = list('abc')
        size = 3
        constraints = []

        # Set BQM and constraints of varying lengths
        for triplet in itertools.combinations(variables, size):
            for u, v in itertools.combinations(triplet, 2):
                bqm.add_interaction(u, v, -1)
            non_fixed_variables = set(triplet) - set(fixed_variables)
            constraints.append(non_fixed_variables)

        for fixed_variable in fixed_variables:
            bqm.fix_variable(fixed_variable, 1)

        # Get new state
        rcd = RandomConstraintDecomposer(size, constraints)
        state = State.from_sample(min_sample(bqm), bqm)
        newstate = rcd.run(state).result()

        self.assertIn('subproblem', newstate)
        self.assertTrue(len(newstate.subproblem) <= size)  # correct size

    def test_partially_disconnected_constraints(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': 1, 'bc': 1, 'ca': 1})
        size = 2
        constraints = ['a', 'b', 'cb']

        rcd = RandomConstraintDecomposer(size, constraints)
        state = State.from_problem(bqm)
        newstate = rcd.run(state).result()

        self.assertIn('subproblem', newstate)
        self.assertTrue(len(newstate.subproblem) <= size)  # correct size

    def test_completely_disconnected_constraints(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({}, {'ab': 1, 'bc': 1, 'ca': 1})
        size = 2
        constraints = ['a', 'b']

        with self.assertRaises(ValueError):
            rcd = RandomConstraintDecomposer(size, constraints)
            state = State.from_problem(bqm)
            _ = rcd.run(state).result()


class TestRoofDualityDecomposer(unittest.TestCase):
    def test_allfixed(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({'a': -1}, {})

        init = State.from_sample({'a': -1}, bqm)  # should be flipped
        new = RoofDualityDecomposer().run(init).result()

        self.assertEqual(new.problem, bqm)
        self.assertEqual(new.subproblem,
                         dimod.BinaryQuadraticModel.from_ising({}, {}, -1))
        self.assertEqual(new.samples.record.sample, [1])

    def test_sampling_mode(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({'a': 0}, {})

        init = State.from_samples([{'a': -1}, {'a': 1}], bqm)

        # variable not fixed
        new = RoofDualityDecomposer(sampling_mode=True).run(init).result()
        self.assertEqual(new.problem, bqm)
        self.assertEqual(new.subproblem, bqm)
        self.assertEqual(new.samples, init.samples)

        # variable fixed
        new = RoofDualityDecomposer(sampling_mode=False).run(init).result()
        self.assertEqual(new.problem, bqm)
        self.assertEqual(new.subproblem,
                         dimod.BinaryQuadraticModel.from_ising({}, {}))
        self.assertTrue(len(set(new.samples.record.sample.flatten())), 1)

    def test_energy_correctness(self):
        bqm = dimod.BinaryQuadraticModel.from_ising({'a': -1}, {})

        init = State.from_sample({'a': -1}, bqm)  # should be flipped
        new = RoofDualityDecomposer().run(init).result()

        self.assertEqual(new.samples.record.energy,
                         bqm.energies((new.samples.record.sample,
                                       new.samples.variables)))

class TestMakeOriginEmbeddings(unittest.TestCase):
    
    def test_default_embeddings(self):
        """Check default workflow, and self-consistency of default dimensions.
        """
        qpu_sampler = MockDWaveSampler()
        orig_embs = make_origin_embeddings(qpu_sampler=qpu_sampler)
        # Non-empty list of non-empty dictionaries
        self.assertTrue(isinstance(orig_embs, list))
        self.assertTrue(len(orig_embs)>0)
        for orig_emb in orig_embs:
            self.assertTrue(isinstance(orig_emb, dict))
            self.assertTrue(len(orig_emb)>0)

        # variable key should be a tuple of consistent length (function of
        # default qpu_solver topology vector coordinate scheme)
        # chain length should be 1 (native default embedding)
        for orig_emb in orig_embs:
            dict_item = next(iter(orig_emb.items()))
            tuple_length = len(dict_item[0])
            chain_length = len(dict_item[1])
        
            for key, val in orig_emb.items():
                self.assertEqual(len(key), tuple_length)
                self.assertEqual(len(val), chain_length)
  
    def test_all_embedding_shapes(self):
        """Check embeddings match anticipated optimal dimensions for Chimera,
        Pegasus and Zephyr processors with Cubic, Kings and Native embeddings.
        Uses a default processor scale of 4, with either 0 or 15
        edge defects.
        """
        # Expected properties by (topology_type, lattice_topology):
        # tuple length, chain length, number of embeddings
        shape_dicts = {('zephyr', 'zephyr'): {'tl': 4, 'cl': 1, 'ne': 2},
                       ('zephyr', 'kings'): {'tl': 2, 'cl': 2, 'ne': 2},
                       ('zephyr', 'cubic'): {'tl': 3, 'cl': 2, 'ne': 3},
                       ('pegasus', 'pegasus'): {'tl': 5, 'cl': 1, 'ne': 2},
                       ('pegasus', 'cubic'): {'tl': 3, 'cl': 2, 'ne': 3},
                       ('pegasus', 'kings'): {'tl': 2, 'cl': 2, 'ne': 2},
                       ('chimera', 'chimera'): {'tl': 4, 'cl': 1, 'ne': 2},
                       ('chimera', 'cubic'): {'tl': 3, 'cl': 4, 'ne': 3}}
        
        for qpu_top in ['pegasus', 'chimera', 'zephyr']:
            if qpu_top == 'chimera':
                lattice_types = ['cubic', qpu_top, None]
            elif qpu_top == 'pegasus':
                lattice_types = ['cubic', 'kings', qpu_top, None]
            elif qpu_top == 'zephyr':
                lattice_types = ['cubic', 'kings', qpu_top, None]
            else:
                raise ValueError('Unknown qpu topology')
            #Native by default:
            shape_dicts[(qpu_top, None)] = shape_dicts[(qpu_top, qpu_top)] 
            qpu_sampler = MockDWaveSampler(topology_type=qpu_top)

            # pop final 15 edges to exercise edge cover routines.
            # 15 is a worst case upper bound on the number of defects that
            # can be handled given the default exact edge cover routines. 
            for pop_edges in [0, 15]:
                for _ in range(pop_edges):
                    qpu_sampler.edgelist.pop()
                qpu_sampler.properties['couplers'] = qpu_sampler.edgelist

                for lattice_type in lattice_types:
                    orig_embs = make_origin_embeddings(
                        qpu_sampler=qpu_sampler,
                        lattice_type=lattice_type)
                    shape_dict = shape_dicts[(qpu_top, lattice_type)]
                    tuple_length = shape_dict['tl'] 
                    chain_length = shape_dict['cl']
                    self.assertEqual(len(orig_embs), shape_dict['ne'])
                    for orig_emb in orig_embs:
                        for key, val in orig_emb.items():
                            self.assertEqual(len(key), tuple_length)
                            self.assertEqual(len(val), chain_length)

    def test_make_cubic_lattice(self):
        dims = (5,3,4)
        g = _make_cubic_lattice(dims)
        self.assertEqual(g.number_of_nodes(), dims[0]*dims[1]*dims[2])
        self.assertEqual(g.number_of_edges(),
                         (dims[0]-1)*dims[1]*dims[2] +
                         dims[0]*(dims[1]-1)*dims[2] +
                         dims[0]*dims[1]*(dims[2]-1))
        for is_open in itertools.product((0, 1), (0, 1), (0, 1)):
            g = _make_cubic_lattice(dims, is_open=is_open)
            self.assertEqual(g.number_of_nodes(), dims[0]*dims[1]*dims[2])
            self.assertEqual(g.number_of_edges(),
                             (dims[0]-is_open[0])*dims[1]*dims[2] +
                             dims[0]*(dims[1]-is_open[1])*dims[2] +
                             dims[0]*dims[1]*(dims[2]-is_open[2]))
        
    def test_make_kings_lattice(self):
        dims = (5,4)
        g = _make_kings_lattice(dims)
        self.assertEqual(g.number_of_nodes(), dims[0]*dims[1])
        self.assertEqual(g.number_of_edges(),
                         (dims[0]-1)*dims[1] +
                         dims[0]*(dims[1]-1) +
                         2*(dims[0]-1)*(dims[1]-1))
        for is_open in itertools.product((0, 1), (0, 1)):
            g = _make_kings_lattice(dims, is_open=is_open)
            self.assertEqual(g.number_of_nodes(), dims[0]*dims[1])
            self.assertEqual(g.number_of_edges(),
                             (dims[0]-is_open[0])*dims[1] +
                             dims[0]*(dims[1]-is_open[1]) +
                             2*(dims[0]-is_open[0])*(dims[1]-is_open[1]))
        
    def test_all_embeddings_validity(self):
        """Check that embeddings are valid for supported lattice_types.
        Uses a default processor scale of 5, with 15 edge defects.
        """
        # Full scale is 16, a smaller default is used
        qpu_scale = 5
        for qpu_top in ['pegasus', 'chimera', 'zephyr']:
            if qpu_top == 'pegasus':
                qpu_shape = [qpu_scale]
                lattice_types = ['cubic', qpu_top, 'kings']
            elif qpu_top == 'chimera':
                qpu_shape = [qpu_scale, qpu_scale, 4]
                lattice_types = ['cubic', qpu_top]
            elif qpu_top == 'zephyr':
                qpu_shape = [qpu_scale, 4]
                lattice_types = ['cubic', 'kings', qpu_top]

            for lattice_type in lattice_types:
                # proposed_source: a defect free-lattice at sampler
                # scale (hence inclusive of all keys).
                if lattice_type == 'cubic':
                    if qpu_top == 'zephyr':
                        cubic_dims = (qpu_scale, qpu_scale, 16)
                    elif qpu_top == 'pegasus':
                        cubic_dims = (qpu_scale-1, qpu_scale-1, 12)
                    else:
                        cubic_dims = (qpu_scale//2, qpu_scale//2, 8)
                    proposed_source = _make_cubic_lattice(cubic_dims)
                elif lattice_type == 'kings':
                    if qpu_top == 'chimera':
                        continue
                    elif qpu_top == 'pegasus':
                        kings_dims = (3*(qpu_scale-1), 3*(qpu_scale-1))
                    else:
                        kings_dims = (4*qpu_scale, 4*qpu_scale)
                    proposed_source = _make_kings_lattice(kings_dims)
                elif lattice_type == 'zephyr':
                    # Chimeralike system for geometric interpretation
                    proposed_source = dnx.zephyr_graph(qpu_scale, coordinates=True)
                    proposed_source= nx.relabel_nodes(
                        G=proposed_source,
                        mapping={n: _zephyr_to_chimeralike(
                            n, t=qpu_shape[1], half_offset=False) for n in proposed_source})
                elif lattice_type == 'pegasus':
                    proposed_source = dnx.pegasus_graph(qpu_scale,
                                                        nice_coordinates=True)
                else:
                    proposed_source = dnx.chimera_graph(qpu_scale, coordinates=True)

                qpu_sampler = MockDWaveSampler(
                    topology_type=qpu_top,
                    topology_shape=qpu_shape)

                orig_embs = make_origin_embeddings(qpu_sampler=qpu_sampler,
                                                   lattice_type=lattice_type)

                # No defects should be happy with all chains, and all edges,
                # for full graph
                for orig_emb in orig_embs:
                    self.assertEqual(len(list(orig_emb.keys())),
                                     proposed_source.number_of_nodes(),
                                     f'Failed: {qpu_top} {lattice_type}')
                # NB, only emb[0] due to technicality
                orig_emb = orig_embs[0]  
                self.assertTrue(verify_embedding(
                    emb=orig_emb,
                    source=proposed_source,
                    target=qpu_sampler.properties['couplers']),
                                f'{qpu_top} and {lattice_type}')

                # With defects, should be happy on subgraph
                for _ in range(15):
                    qpu_sampler.edgelist.pop()
                orig_embs = make_origin_embeddings(qpu_sampler=qpu_sampler,
                                                   lattice_type=lattice_type)
                for orig_emb in orig_embs:
                    self.assertTrue(verify_embedding(
                        emb=orig_emb,
                        source=proposed_source.subgraph(list(orig_emb.keys())),
                        target=qpu_sampler.properties['couplers']))

    def test_chimeralike_coordinates(self):
        known_mappings = {False: {(0,0,u,t): (u,0,t,0,0)
                              for u in range(2) for t in range(4)},
                          True: {(0,0,0,0): (0,0,2,0,0),
                                 (0,0,0,3): (0,1,1,0,0),
                                 (0,0,1,0): (1,0,2,0,0),
                                 (0,0,1,3): (1,1,1,0,0),
                                 (1,0,0,0): (0,0,2,1,0),
                                 (0,1,0,0): (0,1,2,0,0),
                                 (1,1,1,2): (1,2,0,1,0),
                                 (1,2,0,1): (0,2,3,1,0)}}
        for half_offset, known_mapping in known_mappings.items():
            for k, v in known_mapping.items():
                mapped_k = _chimeralike_to_zephyr(k, half_offset=half_offset)
                self.assertEqual(mapped_k, v)
                mapped_v = _zephyr_to_chimeralike(v, half_offset=half_offset)
                self.assertEqual(mapped_v, k)

    def test_constrained_validity(self):
        """Check that we can constrain an embedding to a given subspace
        """
        # Full scale is 16, a smaller default is used
        qpu_scale = 5
        constrained_scales = {'cubic': (2,2,2),
                              'pegasus': (3,2,3,2,4),  # 2x3 nice-cells
                              'chimera': (1,1,2,4),  # single cell
                              'kings': (2,2)  # single cell
        }
        for qpu_top in ['pegasus', 'chimera', 'zephyr']:
            lattice_types = ['cubic']
            if qpu_top != 'zephyr':
                lattice_types += [qpu_top]
            if qpu_top != 'chimera':
                lattice_types.append('kings')

            if qpu_top == 'pegasus':
                qpu_shape = [qpu_scale]
            elif qpu_top == 'chimera':
                qpu_shape = [qpu_scale, qpu_scale, 4]
            elif qpu_top == 'zephyr':
                qpu_shape = [qpu_scale, 4]

            for lattice_type in lattice_types:
                # proposed_source: a defect free-lattice at sampler
                # scale (hence inclusive of all keys).
                qpu_sampler = MockDWaveSampler(
                    topology_type=qpu_top,
                    topology_shape=qpu_shape)
                cs = constrained_scales[lattice_type]
                orig_embs = make_origin_embeddings(qpu_sampler=qpu_sampler,
                                                   lattice_type=lattice_type,
                                                   problem_dims=cs,
                                                   reject_small_problems=False)
                for orig_emb in orig_embs:
                    from numpy import prod
                    self.assertEqual(len(orig_emb), prod(cs))
                    self.assertFalse(any(any(key[idx] >= bound for idx,bound in enumerate(cs)) for key in orig_emb))



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Greedy large neighborhood local search workflows for lattices."""

import hybrid
import dimod
from dwave.system import DWaveSampler

__all__ = ['LatticeLNLS','LatticeLNLSSampler']


def LatticeLNLS(topology,
                exclude_dims=None,
                qpu_sampler=None,
                energy_threshold=None,
                max_iter=128,
                max_time=None,
                convergence=None,
                qpu_params=None,
                workflow_type='qpu-only',
                track_qpu_branch=False):
    '''Implements lattice workflows as described in `Hybrid quantum annealing for
    larger-than-QPU lattice-structured problems <https://arxiv.org/abs/2202.03044>`_.

    LatticeLNLS workflow is used by :class:`LatticeLNLSSampler`.
    Note that to operate this workflow a minimal set of lattice specific state
    variables must be instantiated:

    1. bqm: A :obj:`~dimod.BinaryQuadraticModel`, with variables indexed geometrically

    2. origin_embeddings: see :class:`~hybrid.decomposers.make_origin_embeddings`

    3. problem_dims: see :class:`~hybrid.decomposers.SublatticeDecomposer`

    Args:

        topology (str):
            A lattice topology (e.g. 'cubic'), consistent with bqm.

            Supported values:

                * 'zephyr' (``qpu_sampler`` must be zephyr-structured)

                * 'pegasus' (``qpu_sampler`` must be pegasus-structured)

                * 'cubic' (``qpu_sampler`` must be zephyr, pegasus of chimera-structured)

                * 'kings' (``qpu_sampler`` must be zephyr or pegasus-structured)

                * 'chimera' (``qpu_sampler`` must be chimera-structured)

        qpu_sampler (:class:`dimod.Sampler`, optional, default=DWaveSampler()):
            Sampler such as a D-Wave system.

        qpu_params (dict, optional, default = ``{'num_reads': 25, 'annealing_time': 100}``):
            Dictionary of keyword arguments with values that will be used
            on every call of the QPU sampler.
            A local copy of the parameter is made. If the dictionary does not
            include 'num_reads', it is defaulted as 25, if dictionary
            does not include 'annealing_time', it is defaulted as 100.

        workflow_type (str, optional):
            Options are:

               * 'qpu-only'
                   Default workflow of this paper

               * 'qpu+post-process'
                   Steepest greedy descent over the subspace is run
                   sequentially on samples returned by the QPU.

               * 'qpu+parallel-process'
                   Steepest greedy descent on the full space is run in
                   parallel with the QPU; the best result is accepted on
                   each iteration.

        track_qpu_branch (bool, optional, default=False):
            Flag indicating whether to track samples, subproblems, and subsamples (samples for the
            subproblem) in the `hybrid.State` of the workflow.

        max_iter (int, optional, default=128):
            Number of iterations in the hybrid algorithm.
            Applied per sample.

        max_time (float/None, optional):
            Wall clock runtime termination criterion. Unlimited by default.
            Applied per sample.

        convergence (int, optional):
            Number of iterations with no improvement that terminates sampling.
            Applied per sample.

        energy_threshold (float, optional):
            Terminate when this energy threshold is surpassed. Check is
            performed at the end of each iteration. Applied per sample.

    Returns:
        Workflow (:class:`~hybrid.core.Runnable` instance).

    See also:
        :class:`~hybrid.decomposers.make_origin_embeddings`

        :class:`~hybrid.decomposers.SublatticeDecomposer`

        Jack Raymond et al, `Hybrid quantum annealing for larger-than-QPU
        lattice-structured problems <https://arxiv.org/abs/2202.03044>`_
    '''
    if exclude_dims is None:
        exclude_dims = []
    if qpu_params is None:
        qpu_params = {'num_reads': 25, 'annealing_time': 100}
    if qpu_sampler is None:
        qpu_sampler = DWaveSampler()
    qpu_params0 = qpu_params.copy()
    if 'num_reads' not in qpu_params0:
        qpu_params0['num_reads'] = 25
    if 'annealing_time' not in qpu_params0:
        qpu_params0['annealing_time'] = 100


    def update_tracking_data(lambda_next, state: hybrid.State):
        updates = dict()
        for k1, k2 in [('tracked_samples', 'samples'),
                       ('tracked_subsamples', 'subsamples'),
                       ('tracked_subproblems', 'subproblem')]:
            updates[k1] = state.get(k1, []) + [state.get(k2)]
        state = state.updated(**updates)
        return state

    qpu_branch = (hybrid.decomposers.SublatticeDecomposer()
                  | hybrid.QPUSubproblemExternalEmbeddingSampler(
                      qpu_sampler=qpu_sampler,
                      sampling_params=qpu_params0,
                      num_reads=qpu_params0['num_reads']))

    if workflow_type == 'qpu-only':
        per_it_runnable =  (qpu_branch| hybrid.SplatComposer())
    elif workflow_type == 'qpu+post-process':
        per_it_runnable = (qpu_branch
                           | hybrid.SteepestDescentSubProblemSampler()
                           | hybrid.SplatComposer())
    elif workflow_type == 'qpu+parallel-process':
        if track_qpu_branch:
            raise NotImplementedError(
                'Tracking qpu branch for the "qpu+parallel-process" workflow is not implemented'
                )
        per_it_runnable = (
            hybrid.Parallel(
                qpu_branch | hybrid.SplatComposer(),
                hybrid.SteepestDescentProblemSampler())
            | hybrid.ArgMin())
    else:
        raise ValueError('Unknown workflow type')
    if track_qpu_branch:
        per_it_runnable = per_it_runnable | hybrid.Lambda(update_tracking_data)
    if energy_threshold is not None:
        energy_reached = lambda en: en <= energy_threshold
    else:
        energy_reached = None
    #Iterate to a termination criteria, integrate proposal if energy lowered:
    workflow = hybrid.Loop(per_it_runnable | hybrid.TrackMin(output=True),
                           max_iter=max_iter, terminate=energy_reached,
                           convergence=convergence, max_time=max_time)
    return workflow


class LatticeLNLSSampler(dimod.Sampler):
    """A dimod-compatible hybrid decomposition sampler for problems of lattice
    structure.

    For workflow and lattice related arguments, see:
    :class:`~hybrid.reference.lattice_lnls.LatticeLNLS`.

    Examples:
        This example solves a cubic-structured BQM using the default QPU.
        An 18x18x18 cubic-lattice ferromagnet is created, and sampled
        by the lattice workflow.

        >>> import dimod
        >>> import hybrid
        >>> from dwave.system import DWaveSampler
        >>> topology = 'cubic'
        >>> qpu_sampler = DWaveSampler()                        # doctest: +SKIP
        >>> sampler = hybrid.LatticeLNLSSampler()               # doctest: +SKIP
        >>> cuboid = (18,18,18)
        >>> edge_list = ([((i,j,k),((i+1)%cuboid[0],j,k)) for i in range(cuboid[0])
        ...              for j in  range(cuboid[1]) for k in range(cuboid[2])]
        ...             + [((i,j,k),(i,(j+1)%cuboid[1],k)) for i in range(cuboid[0])
        ...                for j in  range(cuboid[1]) for k in range(cuboid[2])]
        ...             + [((i,j,k),(i,j,(k+1)%cuboid[2])) for i in range(cuboid[0])
        ...                for j in range(cuboid[1]) for k in range(cuboid[2])])
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {e: -1 for e in edge_list})
        >>> EGS = -len(edge_list)
        >>> qpu_params={'num_reads': 25,
        ...             'annealing_time': 100,
        ...             'auto_scale': False,
        ...             'chain_strength': 2}
        >>> response = sampler.sample(topology='cubic', bqm=bqm,
        ...                           problem_dims=cuboid,
        ...                           energy_threshold=EGS,
        ...                           qpu_sampler=qpu_sampler,
        ...                           qpu_params=qpu_params)                           # doctest: +SKIP
        >>> response.data_vectors['energy']                      # doctest: +SKIP
        array([-17496])

    See also:
        :class:`~hybrid.decomposers.make_origin_embeddings`

        :class:`~hybrid.decomposers.SublatticeDecomposer`

        Jack Raymond et al, `Hybrid quantum annealing for larger-than-QPU
        lattice-structured problems <https://arxiv.org/abs/2202.03044>`_
    """

    properties = None
    parameters = None
    runnable = None
    origin_embedding = None

    def __init__(self):
        #Minimum requirements for dimod compatibility are used.
        #Certain parameters might be initialized in principle and
        #shared amongst many sampling processes.
        self.parameters = {
            'origin_embeddings': None
        }
        self.properties = {}

    def sample(self, topology, bqm, problem_dims, exclude_dims=None,
               reject_small_problems=True, qpu_sampler=None,
               init_sample=None, num_reads=1, track_qpu_branch=False, **kwargs):
        """Solve large subspaces of a lattice structured problem sequentially
        integrating proposals greedily to arrive at a global or local minima of
        the bqm.

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from. Keying of variables
                must be appropriate to the lattice class.

            init_sample (:class:`~dimod.SampleSet`, callable, ``None``):
                Initial sample set (or sample generator) used for each "read".
                Use a random sample for each read by default.

            num_reads (int):
                Number of reads. Each sample is the result of a single run of
                the hybrid algorithm.

            track_qpu_branch (bool, optional, default=False):
                Flag indicating whether to track samples, subproblems, and subsamples (samples for
                the subproblem) in the returned sample set's `info` field. One list of tracked data
                is stored per read (as in `num_reads`).

            problem_dims (tuple of ints):
                Lattice dimensions (e.g. cubic case (18,18,18)).

            exclude_dims (list of ints, optional):
                Subspaces are selected by geometric displacement. In the case of
                cellular-level displacements only dimensions indexing cell-displacements
                are considered. The defaults are topology dependent:

                * 'chimera': [2] (u chimera coordinates are not displaced).

                * 'pegasus': [0,3,4] (t,u,k nice pegasus coordinates are not displaced).
                * 'zephyr': [2] (u chimera-like coordinates are not displaced).
                * 'cubic': [] all dimensions are displaced.

            reject_small_problems (bool, optional, default=True):
                If the subsolver is bigger than the target problem, raise an
                error by default (True), otherwise quietly shrink the embedding
                to be no larger than the target problem.

            additional workflow arguments:
                per :class:`~hybrid.reference.lattice_lnls.LatticeLNLS`.

        Returns:
            :class:`~dimod.SampleSet`: A `dimod` :class:`.~dimod.SampleSet` object.

        See also:
            :class:`~hybrid.decomposers.make_origin_embeddings`

            :class:`~hybrid.decomposers.SublatticeDecomposer`

            Jack Raymond et al, `Hybrid quantum annealing for larger-than-QPU
            lattice-structured problems <https://arxiv.org/abs/2202.03044>`_
        """
        if qpu_sampler is None:
            qpu_sampler = DWaveSampler()

        if exclude_dims is None:
            if topology == 'chimera':
                exclude_dims = [2]
            elif topology == 'pegasus':
                exclude_dims = [0,3,4]
            elif topology == 'zephyr':
                exclude_dims = [2]
            else:
                exclude_dims = []
                #Recreate on each call, no reuse:
        self.origin_embeddings = hybrid.make_origin_embeddings(
            qpu_sampler, topology, problem_dims=problem_dims,
            reject_small_problems=reject_small_problems)

        if callable(init_sample):
            init_state_gen = lambda: hybrid.State.from_sample(
                init_sample(),
                bqm,
                problem_dims=problem_dims,
                exclude_dims=exclude_dims,
                origin_embeddings=self.origin_embeddings)
        elif init_sample is None:
            init_state_gen = lambda: hybrid.State.from_sample(
                hybrid.random_sample(bqm),
                bqm,
                problem_dims=problem_dims,
                exclude_dims=exclude_dims,
                origin_embeddings=self.origin_embeddings)
        elif isinstance(init_sample, dimod.SampleSet):
            init_state_gen = lambda: hybrid.State.from_sample(
                init_sample,
                bqm,
                problem_dims=problem_dims,
                exclude_dims=exclude_dims,
                origin_embeddings=self.origin_embeddings)
        else:
            raise TypeError("'init_sample' should be a SampleSet or a SampleSet generator")


        samples = []
        energies = []
        if track_qpu_branch:
            info = dict(tracked_samples=[],
                        tracked_subsamples=[],
                        tracked_subproblems=[])
        else:
            info = dict()
        for _ in range(num_reads):
            #Recreate on each call, no reuse:
            self.runnable = LatticeLNLS(topology=topology,
                                        qpu_sampler=qpu_sampler,
                                        exclude_dims=exclude_dims,
                                        track_qpu_branch=track_qpu_branch,
                                        **kwargs)  # Must be here, or else remembers best state
            init_state = init_state_gen()
            final_state = self.runnable.run(init_state)
            # the best sample from each run is one "read"
            resolved_state = final_state.result()
            ss = resolved_state.samples
            ss.change_vartype(bqm.vartype, inplace=True)
            samples.append(ss.first.sample)
            energies.append(ss.first.energy)
            if track_qpu_branch:
                info['tracked_samples'].append(resolved_state.tracked_samples)
                info['tracked_subsamples'].append(resolved_state.tracked_subsamples)
                info['tracked_subproblems'].append(resolved_state.tracked_subproblems)
        return dimod.SampleSet.from_samples(samples, vartype=bqm.vartype,
                                            energy=energies, info=info)
